import shapely.geometry
from owslib.wms import WebMapService

from munibot.config import config


class BaseProfile:
    """
    Base profile class that all profile bot classes should inherit from.

    Contains mandatory hooks that all profiles need to implement, optional
    hooks and utility methods.
    """

    # Mandatory properties

    """
    A unique identifier for the profile, that will be used when running the
    commands and on the configuration file.
    """
    id = None

    """
    A longer description of the profile.
    """
    desc = None

    # Optional properties

    """
    Value that NODATA pixels have in the base image. Usually 0 or 255.
    """
    image_nodata_value = 0

    # Mandatory hooks

    """
    Given a feature id, returns a tuple with the extent and the geometry of the
    the boundaries of the feature.

    The extent is a tuple containing (minx, miny, maxx, maxy). The geometry is a
    dict containing the geometry in GeoJSON format.
    """

    def get_boundaries(self, id_):

        raise NotImplementedError

    """
    Returns a base image for the given extent (minx, miny, maxx, maxy).

    The return value is a file-like object containing the raster image for the
    provided extent.
    """

    def get_base_image(self, extent):

        raise NotImplementedError

    """
    Returns the text that needs to be included in the tweet for this particular
    feature.
    """

    def get_text(self, id_):

        raise NotImplementedError

    """
    Returns the id of the next feature that should be tweeted.

    This is used if the ``munibot tweet`` command is called withot providing
    an id.
    """

    def get_next_id(self):

        raise NotImplementedError

    # Optional hooks

    """
    Return a tuple with the longitude and latitude that should be associated
    with the tweet.

    The bot account should have location information on tweets activated.
    """

    def get_lon_lat(self, id_):

        return None, None

    """
    Function that will be called after sending the tweet, that will receive
    the feature and the status id of the tweet sent.
    """

    def after_tweet(self, id_, status_id):

        pass

    # Internal

    def __init__(self):
        if not self.id or not self.desc:
            class_name = self.__class__.__name__
            raise NotImplementedError(
                f'Profile class {class_name} must define the "id" and "desc" properties'
            )

    # Utilities

    """
    Returns the vertical and horizontal distance of the provided bounding box.
    """

    def get_bbox_distances(self, bbox):
        dx = abs(bbox[0] - bbox[2])
        dy = abs(bbox[1] - bbox[3])

        return dx, dy

    """
    Calculates the image size based on the bounding box provided and
    the ``max_pixel_side`` configuration option.
    """

    def get_image_size(self, bbox):

        dx, dy = self.get_bbox_distances(bbox)

        max_pixel_side = int(config["image"]["max_pixel_side"])

        if dx > dy:
            return (max_pixel_side, max_pixel_side * dy / dx)
        else:
            return (max_pixel_side * dx / dy, max_pixel_side)

    """
    Extends the bounding box to include a buffer zone around it.
    """

    def extend_bbox(self, bbox):

        dx, dy = self.get_bbox_distances(bbox)

        distance = dx * 0.1 if dx > dy else dy * 0.1
        return shapely.geometry.box(*bbox).buffer(distance).bounds

    """
    Helper function to request a WMS image.

    Returns a file-like object with the requested image.
    """

    def get_wms_image(
        self,
        url,
        layer,
        bbox,
        crs,
        version="1.3.0",
        styles=["default"],
        format="image/tiff",
    ):

        wms = WebMapService(url, version=version)

        img = wms.getmap(
            layers=[layer],
            srs=crs,
            styles=styles,
            bbox=bbox,
            size=self.get_image_size(bbox),
            format=format,
        )

        return img



import argparse
import sys
from urllib.parse import parse_qs, urlparse

from .config import load_config, load_profiles, get_logger
from .image import create_image
from .tweet import get_verify_auth, send_tweet


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command",
        choices=["tweet", "create", "tokens"],
        help="""Action to perform. \"tweet\" sends out a tweet, \"create\" just generates the image locally. Use \"tokens\" to get the API access tokens for a bot""",
    )

    parser.add_argument(
        "profile",
        help="""Profile to use to generate the image / tweet. Must be one of the ones available""",
    )
    parser.add_argument(
        "--id",
        "-i",
        help="""
        Identifier of the administrative unit to create an image of.
        If not provided the relevant profile will provide one
        (generally a random one)""",
    )
    parser.add_argument(
        "--config",
        "-c",
        default="munibot.ini",
        help="""
    Path to the configuration ini file. Defaults to a "munibot.ini" in the same
    folder the command is run on.
    """,
    )
    parser.add_argument(
        "--output",
        "-o",
        default=None,
        help="""
    Output path for the generated image (Only used with "create"). Defaults to a <id>.jpg file in the current folder.
    """,
    )

    args = parser.parse_args()
    try:
        load_config(args.config)
    except ValueError as e:
        print(str(e))
        sys.exit(1)

    profiles = load_profiles()
    if args.profile not in profiles:
        print(f"Unknown profile: {args.profile}")
        sys.exit(1)

    profile = profiles[args.profile]()

    if args.id:
        id_ = args.id
    else:
        id_ = profile.get_next_id()

    if id_ is None:
        print("No more images to create!")
        sys.exit()

    log = get_logger(__name__)

    if args.command == "create":
        if args.output:
            output = args.output
        else:
            output = f"{id_}.jpg"
        log.info(f"Starts: create image for feature {id_} on profile {args.profile}")
        create_image(profile, id_, output)
    elif args.command == "tweet":

        log.info(f"Start: sending tweet for feature {id_} on profile {args.profile}")
        text = profile.get_text(id_)
        img = create_image(profile, id_)
        lon, lat = profile.get_lon_lat(id_)

        send_tweet(profile, id_, text, img, lon, lat)
    elif args.command == "tokens":
        auth = get_verify_auth()
        verify_url = auth.get_authorization_url()

        oauth_token = parse_qs(urlparse(verify_url).query)["oauth_token"][0]

        msg = f"""
Please visit the following URL logged in as the Twitter bot account for this profile, authorize the application and input the verification code shown.

        {verify_url}

Verification code: """.strip()

        verification_code = input(msg)

        auth.request_token = {
            "oauth_token": oauth_token,
            "oauth_token_secret": verification_code,
        }

        access_token, access_token_secret = auth.get_access_token(verification_code)

        print(
            f"""
Done, access tokens for profile {args.profile}:

twitter_access_token={access_token}
twitter_access_token_secret={access_token_secret}
"""
        )



import io
import urllib
import sqlite3
import xml.etree.ElementTree as ET

import fiona
from owslib.wfs import WebFeatureService

from munibot.config import config

from .base import BaseProfile


class MuniBotEs(BaseProfile):

    id = "es"

    desc = "Municipios España (Ortofoto PNOA)"

    image_nodata_value = 0

    def get_boundaries(self, id_):

        admin_wfs = "https://contenido.ign.es/wfs-inspire/unidades-administrativas"

        wfs = WebFeatureService(admin_wfs, version="2.0.0")

        response = wfs.getfeature(
            storedQueryID="urn:ogc:def:query:OGC-WFS::GetFeatureById",
            storedQueryParams={"ID": "AU_ADMINISTRATIVEUNIT_{}".format(id_)},
        )

        root = ET.fromstring(response.read())
        ns = {"au": "http://inspire.ec.europa.eu/schemas/au/4.0"}

        gml_geometry_element = [e for e in root.find("au:geometry", ns)][0]
        gml_geometry = ET.tostring(gml_geometry_element)

        gml = """
        <gml:FeatureCollection xmlns:gml="http://www.opengis.net/gml/3.2">
            <gml:featureMember>
              <gml:geometry>
                {geometry}
              </gml:geometry>
            </gml:featureMember>
        </gml:FeatureCollection>
        """.format(
            geometry=gml_geometry.decode("utf8")
        ).strip()

        gml_f = io.BytesIO(gml.encode("utf8"))
        with fiona.open(gml_f, "r") as src:
            geometry = [f["geometry"] for f in src][0]
            return src.bounds, geometry

    def get_base_image(self, extent):

        bbox = (extent[1], extent[0], extent[3], extent[2])

        bbox = self.extend_bbox(bbox)

        wms_options = {
            "url": "http://www.ign.es/wms-inspire/pnoa-ma",
            "layer": "OI.OrthoimageCoverage",
            "version": "1.3.0",
            "crs": "EPSG:4258",
            "bbox": bbox,
        }

        return self.get_wms_image(**wms_options)

    def get_text(self, id_):

        db = sqlite3.connect(config["profile:es"]["db_path"])

        data = db.execute(
            """
            SELECT nameunit, nameprov
            FROM munis_esp
            WHERE natcode = ?
            """,
            (id_,),
        )

        name_muni, name_prov = data.fetchone()

        wiki_link = "https://es.wikipedia.org/wiki/{}".format(
            urllib.parse.quote(name_muni.replace(" ", "_"))
        )

        return f"{name_muni} ({name_prov})\n\n\n{wiki_link}"

    def get_next_id(self):

        db = sqlite3.connect(config["profile:es"]["db_path"])

        id_ = db.execute(
            """
            SELECT natcode
            FROM munis_esp
            WHERE tweet_es IS NULL
            ORDER BY RANDOM()
            LIMIT 1"""
        )

        return id_.fetchone()[0]

    def get_lon_lat(self, id_):

        db = sqlite3.connect(config["profile:es"]["db_path"])

        data = db.execute(
            """
            SELECT lon, lat
            FROM munis_esp
            WHERE natcode = ?
            """,
            (id_,),
        )

        lon, lat = data.fetchone()

        return lon, lat

    def after_tweet(self, id_, status_id):

        db = sqlite3.connect(config["profile:es"]["db_path"])

        db.execute(
            """
            UPDATE munis_esp
            SET tweet_es = ?
            WHERE natcode = ?
            """,
            (
                status_id,
                id_,
            ),
        )

        db.commit()



import tempfile

import tweepy

from .config import config, get_logger


def get_verify_auth():
    """
    Return a Tweepy authorization handler instance, with just the api key
    and secret set.

    This is generally used in the ``tokens`` command to request the bot account tokens.
    """
    auth = tweepy.OAuthHandler(
        config["twitter"]["api_key"],
        config["twitter"]["api_key_secret"],
        "oob",
    )

    return auth


def get_auth(profile):
    """
    Returns a Tweepy authorization handler instance using the access tokens for a
    particular profile.

    The returned handled can be used to interact with the Twitter API using the
    profile bot account.
    """

    profile_config = config["profile:" + profile.id]

    auth = tweepy.OAuthHandler(
        config["twitter"]["api_key"],
        config["twitter"]["api_key_secret"],
    )
    auth.set_access_token(
        profile_config["twitter_access_token"],
        profile_config["twitter_access_token_secret"],
    )

    return auth


def send_tweet(profile, id_, text, image, lon=None, lat=None):
    """
    Publish a new tweet in the profile bot account.

    Uploads the provided image and posts a new status with the text provided.
    If provided, it also includes longitude and latitude information, but these
    are optional.

    After sending the tweet it calls the ``after_tweet()`` method of the provided
    profile, passing the feature id and the new tweet status id.

    :param profile: an instance of the profile class
    :type profile: object
    :param id_: the identifier of the feature
    :type id_: string
    :param text: the text to include with the tweet
    :type text: string
    :param image: the image to include with the tweet
    :type image: file-like object
    :param lon: Longitude (optional)
    :type lon: float
    :param lon: Latitude (optional)
    :type lon: float
    """

    log = get_logger(__name__)

    api = tweepy.API(get_auth(profile))

    with tempfile.NamedTemporaryFile(suffix=".jpg") as f:
        f.write(image.getbuffer())
        media = api.media_upload(f.name, file=f)

    status = api.update_status(text, media_ids=[media.media_id], lon=lon, lat=lat)

    profile.after_tweet(id_, status.id)

    log.info(
        f"Tweet sent for feature {id_} on profile {profile.id}, status id: {status.id}"
    )



import sqlite3
import urllib

from munibot.config import config

from .es import MuniBotEs


class MuniBotCat(MuniBotEs):

    id = "cat"

    desc = "Municipis Catalunya (Ortofoto ICGC)"

    image_nodata_value = 255

    def get_base_image(self, extent):

        bbox = (extent[1], extent[0], extent[3], extent[2])

        bbox = self.extend_bbox(bbox)

        wms_options = {
            "url": "https://geoserveis.icgc.cat/icc_ortohistorica/wms/service",
            "layer": "orto5m2016",
            "version": "1.1.1",
            "crs": "EPSG:4258",
            "bbox": bbox,
        }

        return self.get_wms_image(**wms_options)

    def get_next_id(self):

        db = sqlite3.connect(config["profile:cat"]["db_path"])

        id_ = db.execute(
            """
            SELECT natcode
            FROM munis_esp
            WHERE tweet_cat IS NULL
                AND codcomuni = '09'
            ORDER BY RANDOM()
            LIMIT 1"""
        )

        return id_.fetchone()[0]

    def get_text(self, id_):

        db = sqlite3.connect(config["profile:cat"]["db_path"])

        data = db.execute(
            """
            SELECT nameunit, namecomar
            FROM munis_esp
            WHERE natcode = ?
            """,
            (id_,),
        )

        name_muni, name_comarca = data.fetchone()

        wiki_link = "https://ca.wikipedia.org/wiki/{}".format(
            urllib.parse.quote(name_muni.replace(" ", "_"))
        )

        return f"{name_muni} ({name_comarca})\n\n\n{wiki_link}"

    def after_tweet(self, id_, status_id):

        db = sqlite3.connect(config["profile:cat"]["db_path"])

        db.execute(
            """
            UPDATE munis_esp
            SET tweet_cat = ?
            WHERE natcode = ?
            """,
            (
                status_id,
                id_,
            ),
        )

        db.commit()



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    install_requires = f.readlines()

setuptools.setup(
    name="munibot",
    version="0.0.1",
    author="Adrià Mercader",
    author_email="amercadero@gmail.com",
    description="A Twitter bot that tweets aerial imagery pictures of municipalities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amercader/munibot",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": ["munibot=munibot.munibot:main"],
        "munibot_profiles": [
            "es=munibot.profiles.es:MuniBotEs",
            "cat=munibot.profiles.cat:MuniBotCat",
        ],
    },
)



import io
import time

import numpy
import rasterio
import rasterio.mask
import rasterio.plot
from PIL import Image

from .config import config, get_logger

MASK_OPACITY = 70


def get_mask(base_image, boundaries, nodata_value=0):
    """
    Returns an image mask for the base_image provided in the shape of the
    geometry of the boundaries provided.

    The base image and the boundaries geometry must share the same coordinate
    reference system.

    :param base_image: A file-like object with the image
    :type base_image: File-like object
    :param boundaries: a GeoJSON-like dict with the geometry of the boundaries
    :type boundaries: dict
    :param nodata_value: Numeric value that the base image has for NODATA values.
        Defaults to 0.
    :type nodata_value: int

    :returns: numpy-like array that can be transformed into an image
    """
    with rasterio.open(base_image) as base:
        out_raster, out_transform = rasterio.mask.mask(base, [boundaries])

        out_raster_bool = out_raster == nodata_value

        out_raster_int = out_raster_bool.astype(numpy.uint8)
        out_raster_int = out_raster_int * 255

        out_image_array = rasterio.plot.reshape_as_image(out_raster_int)

        return out_image_array


def process_image(base_image, mask_array, output_format="jpeg"):
    """
    Create the final image including the aerial imagery background with the
    outside of the featured boundaries masked.

    :param base_image: A file-like object with the image
    :type base_image: File-like object
    :param mask_array: numpy-like array that can be transformed into an image
    :type mask_array: array
    :param output_format: Output image format. Defaults to "jpeg".
    :type output_format: string

    :returns: resulting image saved as JPG
    :rtype: file-like object
    """

    mask = Image.fromarray(mask_array)
    mask_opacity = int(config["image"]["opacity"])

    alpha = int(mask_opacity * 255 / 100)
    mask.putalpha(alpha)

    pixels = mask.load()

    for x in range(mask.size[0]):
        for y in range(mask.size[1]):
            r, g, b, a = pixels[x, y]
            pixels[x, y] = (r, g, b, a if r == 255 else 0)

    back = Image.open(base_image)
    back.paste(mask, (0, 0), mask)

    out = io.BytesIO()

    back.save(out, format=output_format)

    return out


def create_image(profile, id_, output=None):
    """
    Creates the image to tweet for the provided profile and id

    :param profile: an instance of the profile class
    :type profile: object
    :param id_: the identifier of the feature to generate the image for
    :type id_: string
    :param output: Optional, image output path if provided the image will be
        saved as a JPG file in that location. If not (the default) a file-like
        object will be returned instead.
    :type output: string

    :returns: a file-like object with the generated image (or None
        if ``output`` is provided)
    :rtype: file-like object
    """

    log = get_logger(__name__)
    start = time.perf_counter()

    extent, boundaries = profile.get_boundaries(id_)
    log.debug("Received boundaries from profile ({})".format(extent))

    base_image = profile.get_base_image(extent)
    log.debug("Received base image from profile")

    nodata_value = getattr(profile, "image_nodata_value", 0)
    mask = get_mask(base_image, boundaries, nodata_value)
    log.debug("Image mask created")

    final_image = process_image(base_image, mask)

    end = time.perf_counter()
    log.info(f"Created image {id_}.jpg in {end - start:0.4f} seconds")

    if output:
        with open(output, "wb") as f:
            f.write(final_image.getbuffer())
    else:
        return final_image



import configparser
import os
import logging

from pkg_resources import iter_entry_points

config = {}


def load_config(path):

    if not os.path.exists(path):
        raise ValueError(
            """
INI file not found. It must be a "munibot.ini" file in the current directory,
otherwise pass the location with the "--config" parameter""".strip()
        )

    cp = configparser.RawConfigParser()
    cp.read(path)
    for section in cp.sections():
        config[section] = dict(cp[section])


def load_profiles():

    profiles = {
        profile.name: profile.load()
        for profile in iter_entry_points("munibot_profiles")
    }

    return profiles


def get_logger(name):

    logger = logging.getLogger(name)
    if "logging" in config:
        level = config["logging"].get("level", "WARNING")
        level = getattr(logging, level.upper(), 30)
        logger.setLevel(level)
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter(
            config["logging"].get(
                "format", "%(asctime)s %(levelname)-5.5s [%(name)s] %(message)s"
            )
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger
